/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: Wavelib.h                                                                                      |
|                Author: Michael Krmer                                                                                 |
|                        Michael@dk8pp.de               (private)                                                       |
|                  Date: 20-Apr-2005                                                                                    |
|              Language: Open Watcom C 1.6                                                                              |
|               Version: 1.0                                                                                            |
|        Program status: First public release                                                                           |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Include file for Wavelib functions                                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef WAVELIB
#include <stdint.h>
#include <stdbool.h>

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define WAVELIB 1
#define Mode_Read  0            /* file is open for read */
#define Mode_Write 1            /* file is open for write */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
typedef struct {
        FILE *cfileptr ;
        uint16_t BitsPerSample ;
        uint16_t Channels ;
        uint32_t SamplesPerSecond ;
        uint32_t TotalDataChunkLength ;
        uint32_t RemainDataChunkLength ;
        uint16_t mode ;
                } WFILE ;

typedef struct {
        const char id[4] ;
        uint32_t len ;
        const char wave_id[4] ;
                } RIFF ;

typedef struct {
        const char id[4] ;
        uint32_t len ;
                } CHUNK_HDR ;

typedef struct {
        uint16_t wFormatTag ;
        uint16_t wChannels ;
        uint32_t  dwSamplesPerSecond ;
        uint32_t  dwAvgBytesPerSecond ;
        uint16_t wBlockAlign ;
        uint16_t wBitsPerSample ;
                } FMT_CK ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
WFILE *wfopen(const char *filename) ;
void *wfgetsample(void *sample, WFILE *wstream) ;
void *wfputsample(void *sample, WFILE *wstream) ;
WFILE *wfcreate(const char *filename, uint16_t bps, uint16_t chs, uint32_t sps) ;
void wfclose(WFILE *wstream) ;
int wfeof(WFILE *wstream) ;

#endif
